"""Lexer for VivJson

- Lexer: Constructor. Its argument is source code as string.
- Lexer#get_token: Extract a token from the current position of source
                   code.

For example, the given source code is extracted as below.

              GT  NUMBER  ASSIGN  PLUS  NUMBER
               |  |            |   |    |
    while  ( i >  3     )  { i = i +    1     }
    |      | |          |  | |   |            |
IDENTIFIER | IDENTIFIER |  | IDENTIFIER       |
           |            |  |                  |
       LEFT_PARENTHESIS |  LEFT_CURLY_BRACKET |
                        |                     |
          RIGHT_PARENTHESIS            RIGHT_CURLY_BRACKET


Refer to:
- "Let's make a Teeny Tiny compiler"
  https://austinhenley.com/blog/teenytinycompiler1.html
- "Crafting Interpreters"
  https://craftinginterpreters.com/
Note that this code is made from scratch. The source code
of the above WEB sites is not used.

Environment:
- Python 3.9 or later

License:
Copyright 2025 benesult

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

__author__ = 'Fumiaki Motegi <motegi@benesult.com>'
__date__ = '2025-03-20'

# pylint: disable=R0902, R0903

import re
from .error import Error
from .tokens import Token

class Lexer:
    """Lexer class

    Attributes:
        _medium (str or NoneType): The name of object that has source
                                   code. It is used to report error.
                                   For example, "test.viv",
                                   "1st argument".
        _config (Config or NoneType): Configuration
        _maybe_get_token_methods (tuple): Methods of getting
                                          the particular token.
        _source_code (str): Source code as text. It is given from
                            caller.
        _current_letter (str): A current letter in the text.
        _current_position (int): The current position in the text.
        _token_head_position (int): The position of the current
                                    token's head in the text.
        _line_number (int): Line number in the text. It is
                            saved into Token and is used for
                            error information.
        _line_head_position (int): The position of the current
                                   line's head in the text.
                                   It is used to calculate column
                                   number that is saved into Token.
                                   The column number is used for
                                   error information.
        _previous_token (Token or NoneType): Previous token.
                                             None if first parsing.
        _hex_pattern (re.Pattern): Compiled pattern of Hex-digit.
        _EOS (str): A marker for end of string.
                    It indicates that all letters are consumed.
        _ONE_LETTER_OPERATORS (tuple): Token types of
                                       1 letter's operations
        _TWO_LETTERS_OPERATORS (tuple): 2 dimensional tuple of
                                        Token type.
                                        1st value is Token type of
                                        2 letter's operation.
                                        2nd value is Token type of
                                        first one letter.
        _TERMINATORS (dict): Token types of terminator.
        _WHITE_SPACES (str): White spaces that should be skipped.
        _REPLACED_LETTERS (dict): The escaped letter and real letter.
    """
    _EOS = '\0'
    _ONE_LETTER_OPERATORS = (
        Token.Type.DOT,
        Token.Type.LEFT_PARENTHESIS,
        Token.Type.RIGHT_PARENTHESIS,
        Token.Type.LEFT_SQUARE_BRACKET,
        Token.Type.RIGHT_SQUARE_BRACKET,
        Token.Type.LEFT_CURLY_BRACKET,
        Token.Type.RIGHT_CURLY_BRACKET
    )
    _TWO_LETTERS_OPERATORS = (
        # (for 2 letters, for first 1 letter)
        (Token.Type.EQ, Token.Type.ASSIGN),  # ==, =
        (Token.Type.LE, Token.Type.LT),  # <=, <
        (Token.Type.GE, Token.Type.GT),  # >=, >
        (Token.Type.NE, Token.Type.ERROR),  # !=, (unexpected)
        (Token.Type.PLUS_ASSIGN, Token.Type.PLUS),  # +=, +
        (Token.Type.MINUS_ASSIGN, Token.Type.MINUS),  # -=, -
        (Token.Type.MULTIPLY_ASSIGN, Token.Type.MULTIPLY),  # *=, *
        (Token.Type.DIVIDE_ASSIGN, Token.Type.DIVIDE),  # /=, /
        (Token.Type.MODULO_ASSIGN, Token.Type.MODULO),  # %=, %
        (Token.Type.RESULT, Token.Type.SET)  # :=, :
    )
    _TERMINATORS = {
        '\n': Token.Type.NEW_LINE,
        ';': Token.Type.SEMICOLON,
        ',': Token.Type.COMMA,
        _EOS: Token.Type.EOS
    }
    _WHITE_SPACES = ' \t\r'  # Space, Tab, and Carriage Return
    _REPLACED_LETTERS = {'b': chr(0x08), 'f': chr(0x0C), 'n': chr(0x0A),
                         'r': chr(0x0D), 't': chr(0x09)}

    def __init__(self, source_code, medium=None, config=None):
        """Initialize class.

        Args:
            source_code (str): Source code as text
            medium (str, optional): The name of object that has source
                                    code. It is used to report error.
                                    For example, "test.viv",
                                    "1st argument".
            config (Config or NoneType): Configuration if needed.
        """
        self._medium = medium
        self._config = config
        self._maybe_get_token_methods = (
            self._maybe_get_1_letter_operation,
            self._maybe_get_2_letters_operation,
            self._maybe_get_string,
            self._maybe_get_number,
            self._maybe_get_identifier_or_keyword,
            self._maybe_get_terminator
        )

        self._source_code = source_code + '\n'  # Add for convenient reason

        self._current_letter = ''
        self._current_position = -1  # It becomes 0 in the following function.
        self._go_to_next_letter()
        self._token_head_position = self._current_position

        self._line_number = 0  # It becomes 1 in the following function.
        self._record_new_line()

        self._previous_token = None

        self._hex_pattern = re.compile(r'[\dA-Fa-f]+')

    def get_token(self):
        """Get the current token.

        Returns:
            Token: The current token.

        Raises:
            LexError: When unknown token is found, it is happen.
        """
        while self._skip_white_space() or self._skip_comment():
            pass

        self._token_head_position = self._current_position
        token = None

        for method in self._maybe_get_token_methods:
            token = method()       # None is returned when this method
            if token is not None:  # can't treat the current lexeme.
                break

        # Error if all methods can't treat the current lexeme.
        if token is None or token.type == Token.Type.ERROR:
            self._abort('Unknown token')

        self._previous_token = token
        return token

    def _record_new_line(self):
        """Record about new line."""
        self._line_number += 1
        self._line_head_position = self._current_position

    def _get_column_number(self):
        """Calculate column number.

        Returns:
            int: The column number. Its minimum value is 1.
        """
        return self._token_head_position - self._line_head_position + 1

    def _go_to(self, offset):
        """Seek and get a letter.

        The current position is moved to the given offset.
        Then the current letter is updated. It becomes EOS marker
        when the position is larger than source code's size. In
        other words, it indicates that all letters are consumed.

        Args:
            offset (int): The offset of position in the text.
        """
        self._current_position = \
            max(self._current_position + offset, 0)  # >= 0
        self._current_letter = self._get_letter()

    def _go_to_next_letter(self):
        """Process a next letter.

        The current position is moved to next letter.
        Then the current letter is updated. It becomes EOS marker
        when all letters are consumed.
        """
        self._go_to(1)

    def _peek_next_letter(self):
        """Peek a next letter.

        A next letter is returned without increment of current
        position.

        Returns:
            str: A next letter
                 EOS if the given position is passed through
                 end of text.
        """
        return self._get_letter(1)

    def _get_letter(self, offset=0):
        """Get a letter of the given offset.

        Args:
            offset (int, optional): The offset of position in the text.
                                    It must be positive number.
                                    The default is 0.

        Returns:
            str: A letter of the given offset.
                 EOS if the given position is out of range.
        """
        position = self._current_position + offset
        return self._source_code[position] \
                if position < len(self._source_code) \
                else self._EOS

    def _maybe_get_1_letter_operation(self):
        """Try to get a token of 1 letter's operation.

        Returns:
            Token or NoneType: A token of 1 letter's operation
                               if it is existed in the current
                               position.
                               Otherwise, None is returned.

        Note:
            When Token is returned, the current position & letter
            are progressed. Otherwise, they are kept.
        """
        token = None
        for token_type in self._ONE_LETTER_OPERATORS:
            if self._current_letter == token_type.value:
                token = self._make_token(token_type)
                self._go_to_next_letter()
                break
        return token

    def _maybe_get_2_letters_operation(self):
        """Try to get a token of two letters operation.

        Returns:
            Token or NoneType: A token of "==", "=", "<=", "<",
                               ">=", ">", or "!=" if it is existed
                               in the current position.
                               Otherwise, None is returned.

        Raises:
            LexError: When only "!" is found, it is happen.

        Note:
            When Token is returned, the current position & letter
            are progressed. Otherwise, they are kept.
        """
        token = None
        for token_types in self._TWO_LETTERS_OPERATORS:
            operator = token_types[0].value
            if self._current_letter == operator[0]:
                length = 1
                if self._peek_next_letter() == operator[1]:
                    length = 2
                token_type = token_types[2 - length]
                operator = operator[0:length]
                if token_type == Token.Type.ERROR:
                    self._abort(f'Operation "{operator}" is not allowed here')
                token = self._make_token(token_type, operator)
                self._go_to(length)
                break
        return token

    def _maybe_get_string(self):
        """Try to get a token of string.

        string is any characters as UTF-8. It is surrounded with
        "" or ''.

        Returns:
            Token or NoneType: A token of string if it is existed
                               in the current position.
                               Otherwise, None is returned.

        Raises:
            LexError: When the end of quotation mark is missing,
                      it is happen.

        Note:
            When Token is returned, the current position & letter
            are progressed. Otherwise, they are kept.
        """
        # Need a quotation mark for start.
        if self._current_letter not in ('"', "'"):
            return None
        surround_mark = self._current_letter

        # Remove the above quotation mark.
        self._go_to_next_letter()
        lexeme = ''

        # Loop until reaching a quotation mark or end of string.
        while self._current_letter not in (surround_mark, '\n', self._EOS):
            # Escape sequence
            if self._current_letter == '\\':
                next_letter = self._peek_next_letter()
                # 4 hexdigits unicode
                if next_letter == 'u':
                    # No problem even if the following indexes are
                    # out of range.
                    hexdigits = self._source_code[self._current_position+2:
                                                  self._current_position+6]
                    if len(hexdigits) != 4:
                        self._abort("4 hexdigits unicode is invalid.")
                    re_obj = self._hex_pattern.fullmatch(hexdigits)
                    if re_obj is None:
                        self._abort("4 hexdigits unicode is invalid.")
                    letter = self._source_code[self._current_position:
                                               self._current_position+6]
                    lexeme += letter.encode().decode('unicode_escape')
                    self._go_to(6)
                    continue

                # Quotation mark, Reverse solidus, Solidus
                if next_letter in ('"', "'", '\\', '/'):
                    lexeme += next_letter
                    self._go_to(2)
                    continue

                # BS, Form-feed, LF, CR, Tab
                if next_letter in self._REPLACED_LETTERS:
                    lexeme += self._REPLACED_LETTERS[next_letter]
                    self._go_to(2)
                    continue

            lexeme += self._current_letter
            # Progress
            self._go_to_next_letter()

        if self._current_letter != surround_mark:
            self._abort('Missing the end of quotation mark')

        # Remove the quotation mark.
        self._go_to_next_letter()

        return self._make_token(Token.Type.STRING, lexeme)

    def _maybe_get_number(self):
        """Try to get a token of number.

        Get a token of number as the following format.
        As you can see, the head of sign (+/-) is not given in order to
        parse later.
        For example, 2, 5.7, 100.5e30, 0.8e+2, 4E-100.

        When previous token is "." and "." exists after number, this
        number is treated as int. To tell the truth, this judgement is
        too much as role of Lexer. But it is needed here.

        <number> ::= <digit>+ [ "." <digit>+ ]
            [ ( "e" | "E" ) [ "+" | "-" ] <digit>+ ]
        <digit> ::= "0" | "1" | "2" | "3" | "4" | "5" | "6" | "7" | "8"
            | "9"

        Returns:
            Token or NoneType: A token of number if it is existed
                               in the current position.
                               Otherwise, None is returned.

        Raises:
            LexError: When there is no digit after decimal point,
                      it is happen.

        Note:
            When Token is returned, the current position & letter
            are progressed. Otherwise, they are kept.
        """
        start_position = self._current_position

        is_skipped = self._skip_digit()  # Check and proceed for 0-9.
        if not is_skipped:
            return None

        token_type = Token.Type.NUMBER

        if self._current_letter == '.':  # Check a decimal point.
            # For access of array/block's member with "." operator.
            # For example, the above skipped digits are 10 and the
            # current letter is 2nd dot in "foo.10.3".
            # Avoid to confuse as floating-number.
            if self._previous_token is not None \
                    and self._previous_token.type == Token.Type.DOT:
                lexeme = self._source_code[start_position:
                                           self._current_position]
                return self._make_token(token_type, lexeme)
            # Is there number after dot (decimal point)?
            if self._peek_next_letter().isdigit():
                # For floating-point number
                # For example, the above skipped digits are 10,
                # the current letter is dot, and peeked letter
                # is 3 in "10.35".
                self._go_to_next_letter()  # Skip a decimal point.
                is_skipped = self._skip_digit()  # Check and proceed for 0-9.
                if not is_skipped:
                    self._abort(
                        'Interpreter side error in Lexer#_maybe_get_number')
            else:
                # For access of array/block's member with "." operator.
                # Check the following token may be identifier.
                # Note that white spaces (includes line-break) around dot
                # are permitted, such as "foo.3. bar".
                offset = 1
                while (letter := self._get_letter(offset)) != self._EOS:
                    if letter == '_' or letter.isalpha():
                        lexeme = self._source_code[start_position:
                                                   self._current_position]
                        return self._make_token(token_type, lexeme)
                    if letter != '\n' and letter not in self._WHITE_SPACES:
                        break
                    offset += 1
                self._abort('There is no digit after decimal point')

        if self._current_letter in ('e', 'E'):  # Check a exponential mark.
            self._go_to_next_letter()  # Skip a exponential mark.
            if self._current_letter in ('+', '-'):  # Check a sign.
                self._go_to_next_letter()  # Skip a sign.
            is_skipped = self._skip_digit()  # Check and proceed for 0-9.
            if not is_skipped:
                self._abort('There is no digit after exponential mark')

        # When this line is reached, the current position goes over
        # number. In other words, the current letter is not number.
        lexeme = self._source_code[start_position:self._current_position]

        return self._make_token(token_type, lexeme)

    def _maybe_get_identifier_or_keyword(self):
        """Try to get a token of identifier/keyword.

        | Category          | 1st letter | The following letters |
        |-------------------|------------|-----------------------|
        | Alphabet [a-zA-Z] | Valid      | Valid                 |
        | Digit [0-9]       | Invalid    | Valid                 |
        | Under score "_"   | Valid      | Valid                 |
        | Others            | Invalid    | Invalid               |

        Returns:
            Token or NoneType: A token of identifier/keyword if it is
                               existed in the current position.
                               Otherwise, None is returned.

        Note:
            When Token is returned, the current position & letter
            are progressed. Otherwise, they are kept.
        """
        start_position = self._current_position
        is_processed = False
        while self._current_letter.isalpha() or self._current_letter == '_' \
                or (is_processed and self._current_letter.isdigit()):
            self._go_to_next_letter()
            is_processed = True

        if not is_processed:
            return None

        # When this line is reached, the current position goes over
        # the identifier. In other words, the current letter is not
        # identifier.
        lexeme = self._source_code[start_position:self._current_position]

        token = self._make_token(Token.Type.IDENTIFIER, lexeme)
        token.fix_keyword_type()
        return token

    def _maybe_get_terminator(self):
        """Try to get a token of terminator.

        Returns:
            Token or NoneType: A token of terminator if it is
                               existed in the current position.
                               Otherwise, None is returned.

        Note:
            When Token is returned, the current position & letter
            are progressed. Otherwise, they are kept.
        """
        if self._current_letter not in self._TERMINATORS:
            return None

        token_type = self._TERMINATORS[self._current_letter]
        token = self._make_token(token_type, self._current_letter)
        current_letter = self._current_letter
        self._go_to_next_letter()
        if current_letter == '\n':
            self._record_new_line()
        return token

    def _make_token(self, token_type, lexeme=None):
        """Make a token.

        Args:
            token_type (Token.Type): The type of token
            lexeme (str, optional): The token's actual text. It is so
                                    called "lexeme".
                                    When this is omitted or this is
                                    None, the assigned value of Token
                                    type is used instead it.

        Returns:
            Token: A token.
        """
        return Token(token_type, self._medium,
                     self._line_number, self._get_column_number(),
                     lexeme)

    def _skip_digit(self):
        """Skip digit.

        Returns:
            bool: True if digit is skipped,
                  False otherwise.
        """
        is_skipped = False
        while self._current_letter.isdigit():
            is_skipped = True
            self._go_to_next_letter()
        return is_skipped

    def _skip_white_space(self):
        """Skip white space.

        Skip whitespace except newlines, which we will use to indicate
        the end of a statement.

        Returns:
            bool: True if white spaces are skipped, False otherwise.
        """
        is_skipped = False
        while self._WHITE_SPACES.find(self._current_letter) != -1:
            self._go_to_next_letter()
            is_skipped = True
        return is_skipped

    def _skip_comment(self):
        """Skip comment in the code.

        The following comment is skipped.
        - It is constructed from "#" to the end of line.
        - It is constructed from "//" to the end of line.
        - It is constructed from "/*" to "*/". It can be over multi-line.

        Returns:
            bool: True if comment is skipped, False otherwise.

        Raises:
            LexError: When the end of "*/" is missing, it is happen.
        """
        if (self._current_letter == '/' and self._peek_next_letter() == '/') \
                or self._current_letter == '#':
            while self._current_letter != '\n':
                self._go_to_next_letter()
            return True

        if self._current_letter != '/' or self._peek_next_letter() != '*':
            return False
        self._go_to(2)
        while (self._current_letter != '*' \
               or self._peek_next_letter() != '/'):
            if self._current_letter == self._EOS:
                self._abort('Missing the the end of "*/"')
            current_letter = self._current_letter
            self._go_to_next_letter()
            if current_letter == '\n':
                self._record_new_line()
        self._go_to(2)
        return True

    def _abort(self, message):
        """Abort Lexer.

        Print an error message and raise Exception.

        Args:
            message (str): Error message

        Raises:
            LexError: This exception always happen when this method
                      is called.
        """
        raise Error.make_lex_error(message, self._medium, self._line_number,
                                   self._get_column_number(), self._config)



"""Unit-test for binary expression in Evaluator

Unit-test for "+", "*", "and", "<=", ...

Run `python3 -m pytest tests/test_evaluator_binary.py` in
parent directory.

Environment:
- Python 3.9 or later
- pytest 8.3 or later

License:
Copyright 2025 benesult

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

__author__ = 'Fumiaki Motegi <motegi@benesult.com>'
__date__ = '2025-03-27'

# pylint: disable=C0301, E0401, R0801

import pytest
from vivjson.viv import Viv
from vivjson.config import Config

@pytest.mark.parametrize(
    'prepared_expression, target_expression, expected_value, is_error', (
        # for Evaluator._add
        ('', '{"a": 3, "b": 100} + {"0": "x", "#": true, "/*": false}', {"a": 3, "b": 100, "0": "x", "#": True, "/*": False}, False),  # block + block --> block
        ('', '{"a": 3, "b": 100} + {"0": "x", "#": true, "//": false, "y": -3}', None, True),  # block + block --> ERROR because max array size is limited as 5.
        ('', '{"a": 3, "b": 100} + {"a": -5}', {"a": -2, "b": 100}, False),  # block + block --> block
        ('', '{"a": 3, "b": 100} + [3, 0]', [{"a": 3, "b": 100}, 3, 0], False),  # block + array --> array
        ('', '{"a": 3} + [3, "a"]', [{"a": 3}, 3, "a"], False), # block + array --> array
        ('', '{"a": 3, "b": 100} + "a"', None, True),  # block + string --> ERROR
        ('', '{"a": 3, "b": 100} + 0', None, True),  # block + int --> ERROR
        ('', '{"a": 3, "b": 100} + 0.5', None, True),  # block + float --> ERROR
        ('', '{"a": 3, "b": 100} + false', None, True),  # block + boolean --> ERROR
        ('', '{"a": 3, "b": 100} + null', {"a": 3, "b": 100}, False),  # block + null --> block  (Adding null is meaningless.)
        ('', '[3, "a"] + {"a": 3}', [3, "a", {"a": 3}], False),  # array + block --> array
        ('', '[true, -1] + [+1, false]', [True, -1, [1, False]], False),  # array + array --> array
        ('', '[true, -1, +1, false] + ["a", null]', [True, -1, +1, False, ["a", None]], False),  # array + array --> array
        ('', '[true, -1, +1, false, "a"] + [null]', None, True),  # array + array --> ERROR because max array size is limited as 5.
        ('', '[8.5, null] + "*/"', [8.5, None, "*/"], False),  # array + string --> array
        ('', '[0, false, ] + 8', [0, False, 8], False),  # array + int --> array
        ('', '[0, false, 1, 2] + 8', [0, False, 1, 2, 8], False),  # array + int --> array
        ('', '[0, false, 1, 2, 3] + 8', None, True),  # array + int --> ERROR because max array size is limited as 5.
        ('', '["x", "y"] + 0.000001', ["x", "y", 0.000001], False),  # array + float --> array
        ('', '[3, "a"] + true', [3, "a", True], False),  # array + boolean --> array
        ('', '[0] + null', [0, None], False),  # array + null --> array  (Append null as element even if adding null.)
        ('', '"abc" + {"a": 0}', None, True),  # string + block --> ERROR
        ('', '"abc" + ["def"]', ["abc", "def"], False),  # string + array --> array
        ('', '"abc" + "def"', "abcdef", False),  # string + string --> string
        ('', '"0.0" + 1000', "0.01000", False),  # string + int --> string
        ('', '"88.88" + 1.75', "88.881.75", False),  # string + float --> string
        ('', '"100" + false', "100false", False),  # string + boolean --> string
        ('', '"null" + null', "null", False),  # string + null --> string  (Adding null is meaningless.)
        ('', '10000 + {"a": 0}', None, True),  # int + block --> ERROR
        ('', '10000 + [0]', [10000, 0], False),  # int + array --> array
        ('', '-5 + ".0"', "-5.0", False),  # int + str --> str
        ('', '3 + 2', 5, False),  # int + int --> int
        ('', '-3+-2', -5, False),  # int + int --> int
        ('', '1000000 + 0.1', 1000000.1, False),  # int + float --> float
        ('', '0 + false', False, False),  # int + boolean --> boolean
        ('', '0 + true', True, False),  # int + boolean --> boolean
        ('', '1 + false', True, False),  # int + boolean --> boolean
        ('', '1 + true', True, False),  # int + boolean --> boolean
        ('', '100 + null', 100, False),  # int + null --> int  (Adding null is meaningless.)
        ('', '1.5 + {"x": null}', None, True),  # float + block --> ERROR
        ('', '15e3 + [null, "x"]', [15000.0, None, "x"], False),  # float + array --> array
        ('', '1e-3+"a"', "0.001a", False),  # float + string --> string
        ('', '-0.01e+5+3', -997.0, False),  # float + int --> float
        ('', '5.6 + 8.2', 13.8, False),  # float + float --> float
        ('', '0.0 + false', False, False),  # float + boolean --> boolean
        ('', '0.0 + true', True, False),  # float + boolean --> boolean
        ('', '0.1 + false', True, False),  # float + boolean --> boolean
        ('', '0.1 + true', True, False),  # float + boolean --> boolean
        ('', '1e-10 + null', 1e-10, False),  # float + null --> float  (Adding null is meaningless.)
        ('', 'false + {"a": 3}', None, True),  # boolean + block --> ERROR
        ('', 'false + [3,]', [False, 3], False),  # boolean + array --> array
        ('', 'false + [3,2,1,0]', [False, 3, 2, 1, 0], False),  # boolean + array --> array
        ('', 'false + [4,3,2,1,0]', None, True),  # boolean + array --> ERROR because max array size is limited as 5.
        ('', 'true + "0"', "true0", False),  # boolean + string --> string
        ('', 'false + 0', False, False),  # boolean + int --> boolean
        ('', 'false + 1', True, False),  # boolean + int --> boolean
        ('', 'true + 0', True, False),  # boolean + int --> boolean
        ('', 'true + 1', True, False),  # boolean + int --> boolean
        ('', 'false + 0.0', False, False),  # boolean + float --> boolean
        ('', 'false + 0.1', True, False),  # boolean + float --> boolean
        ('', 'true + 0.0', True, False),  # boolean + float --> boolean
        ('', 'true + 0.1', True, False),  # boolean + float --> boolean
        ('', 'false + false', False, False),  # boolean + boolean --> boolean
        ('', 'false + true', True, False),  # boolean + boolean --> boolean
        ('', 'true + false', True, False),  # boolean + boolean --> boolean
        ('', 'true + true', True, False),  # boolean + boolean --> boolean
        ('', 'true + null', True, False),  # boolean + null --> boolean  (Adding null is meaningless.)
        ('', 'null + {"a": 3}', {"a": 3}, False),  # null + block --> block  (Adding null is meaningless.)
        ('', 'null + [false, [1, 2]]', [None, False, [1, 2]], False),  # null + array --> array  (Append null as element even if adding null.)
        ('', 'null + ""', "", False),  # null + string --> string  (Adding null is meaningless.)
        ('', 'null + "abc"', "abc", False),  # null + string --> string  (Adding null is meaningless.)
        ('', 'null + 0', 0, False),  # null + int --> int  (Adding null is meaningless.)
        ('', 'null + -1e2', -100.0, False),  # null + float --> float  (Adding null is meaningless.)
        ('', 'null + false', False, False),  # null + boolean --> boolean  (Adding null is meaningless.)
        ('', 'null + true', True, False),  # null + boolean --> boolean  (Adding null is meaningless.)
        ('x = len, y = len', 'x + y', None, True),  # function + function --> ERROR
        ('x = len, y = len', '3 + y', None, True),  # int + function --> ERROR
        ('x = len, y = len', 'x + 3', None, True),  # function + int --> ERROR
        ('function test() {return(1)}, x = test(), y = test()', 'x + y', 2, False),
        ('function test() {return(1)}, x = test, y = test', 'x + y', None, True),  # function + function --> ERROR
        ('function test() {return(1)}, x = test(), y = test', 'x + y', None, True),  # int + function --> ERROR
        ('function test() {return(1)}, x = test, y = test()', 'x + y', None, True),  # function + int --> ERROR
        ({"a": 3}, 'a + 2', 5, False),  # int + int --> int
        ({"a": (3,2)}, 'a + 2', None, True),  # Python's tuple can't treated. --> ERROR
                        # When tuple has simple structure, such as 1 dimensional tuple,
                        # converting to list is possible. However complex structure is
                        # given, it is so difficult. So it is not permitted.
        ({"a": NotImplemented}, 'a + 2', None, True),  # Python's NotImplemented can't treated. --> ERROR
        ({"a": Config()}, 'a + 2', None, True),  # Python's class instance can't treated. --> ERROR
        ({"a": "x", "b": "y"}, 'a + b', "xy", False),  # string + string --> string

        # for Evaluator._subtract
        ('', '{"a": 10, "b": 20, "c": 30} - {"b": 5, "c": 10}', {"a": 10, "b": 15, "c": 20}, False),  # block - block --> block
        ('', '{"a": 10, "b": 20} - {"b": 5, "c": 10}', {"a": 10, "b": 15, "c": -10}, False),  # block - block --> block
        ('', '{"a": 10, "b": 20} - {"c": 10, "d": -100, "e": 50, "f": 1}', None, True),  # block - block --> ERROR because max array size is limited as 5.
        ('', '{"a": 10, "b": 20, "c": 30} - ["b", "c"]', {"a": 10}, False),  # block - array --> block
        ('', '{"a": 10, "b": 20, "c": 30} - ["b", "c", 3]', None, True),  # block - array --> ERROR because there is number in array
        ('', '{"a": 10, "b": 20, "c": 30} - [true]', None, True),  # block - array --> ERROR because there is boolean in array
        ('', '{"a": 10, "b": 20, "c": 30} - "b"', {"a": 10, "c": 30}, False),  # block - string --> block
        ('', '{"a": 10, "b": 20, "c": 30} - "d"', {"a": 10, "b": 20, "c": 30}, False),  # block - string --> block
        ('', '{"a": 10} - 3', None, True),  # block - int --> ERROR
        ('', '{"a": 10} - 1.5', None, True),  # block - float --> ERROR
        ('', '{"a": 10} - true', None, True),  # block - boolean --> ERROR
        ('', '{"a": 10} - null', {"a": 10}, False),  # block - null --> block  (Removing null is meaningless.)
        ('', '[3, {"a": null}, {"a": null}] - {"a": null}', [3], False),  # array - block --> array
        ('', '[3, "a", null] - {"a": null}', [3, "a", None], False),  # array - block --> array
        ('', '[3, ["a", null], ["a", null]] - ["a", null]', [3], False),  # array - array --> array
        ('', '[3, ["a", null]] - ["a", null]', [3], False),  # array - array --> array
        ('', '[3, "a", null] - ["a", null]', [3, "a", None], False),  # array - array --> array
        ('', '[3, "a", "a"] - "a"', [3], False),  # array - string --> array
        ('', '[3, "a", "a"] - 3', ["a", "a"], False),  # array - int --> array
        ('', '[3, "a", "a"] - 3.0', ["a", "a"], False),  # array - float --> array
        ('', '[3, true, "a"] - true', [3, "a"], False),  # array - boolean --> array
        ('', '[3, false, "a"] - false', [3, "a"], False),  # array - boolean --> array
        ('', '[3, false, "a"] - true', [3, False, "a"], False),  # array - boolean --> array
        ('', '[3, "a", null] - null', [3, "a"], False),  # array - null --> array  (Remove null of element.)
        ('', '"abc: 3" - {"abc": 3}', None, True),  # string - block --> ERROR
        ('', '"large-dog&small-dog&2cat" - ["large-", "small-"]', "dog&dog&2cat", False),  # string - array --> string
        ('', '"large-dog&small-dog&2cat" - "dog"', "large-&small-&2cat", False),  # string - string --> string
        ('', '"large-dog&small-dog&2cat" - 2', None, True),  # string - int --> ERROR
        ('', '"large-dog&small-dog&200.5t" - 200.5', None, True),  # string - float --> ERROR
        ('', '"large-dog&small-dog&true" - true', None, True),  # string - boolean --> ERROR
        ('', '"large-dog&small-dog&null" - null', "large-dog&small-dog&null", False),  # string - null --> string  (Removing null is meaningless.)
        ('', '100 - {"100": 100}', None, True),  # int - block --> ERROR
        ('', '100 - [100]', None, True),  # int - array --> ERROR
        ('', '100 - "100"', None, True),  # int - string --> ERROR
        ('', '100 - 100', 0, False),  # int - int --> int
        ('', '100 - 100.5', -0.5, False),  # int - float --> float
        ('', '100 - 100.0', 0.0, False),  # int - float --> float
        ('', '100 - true', None, True),  # int - boolean --> ERROR
        ('', '100 - null', 100, False),  # int - null --> int  (Removing null is meaningless.)
        ('', '100.5 - {"100.5": 100.5}', None, True),  # float - block --> ERROR
        ('', '100.5 - [100.5]', None, True),  # float - array --> ERROR
        ('', '100.5 - "100.5"', None, True),  # float - string --> ERROR
        ('', '100.5 - 100', 0.5, False),  # float - int --> float
        ('', '100.5 - 100.5', 0.0, False),  # float - float --> float
        ('', '100.5 - 100.0', 0.5, False),  # float - float --> float
        ('', '100.5 - true', None, True),  # float - boolean --> ERROR
        ('', '100.5 - null', 100.5, False),  # float - null --> float  (Removing null is meaningless.)
        ('', 'true - {"true": true}', None, True),  # boolean - block --> ERROR
        ('', 'true - [true]', None, True),  # boolean - array --> ERROR
        ('', 'true - "true"', None, True),  # boolean - string --> ERROR
        ('', 'true - 1', None, True),  # boolean - int --> ERROR
        ('', 'true - 1.0', None, True),  # boolean - float --> ERROR
        ('', 'true - true', None, True),  # boolean - boolean --> ERROR
        ('', 'true - null', True, False),  # boolean - null --> boolean  (Removing null is meaningless.)
        ('', 'null - {}', None, True),  # null - block --> ERROR
        ('', 'null - {"null": null}', None, True),  # null - block --> ERROR
        ('', 'null - []', None, True),  # null - array --> ERROR
        ('', 'null - [100, 5]', None, True),  # null - array --> ERROR
        ('', 'null - "xyz"', None, True),  # null - string --> ERROR
        ('', 'null - 100', None, True),  # null - int --> ERROR
        ('', 'null - 100.5', None, True),  # null - float --> ERROR
        ('', 'null - true', None, True),  # null - boolean --> ERROR
        ('', 'null - null', None, False),  # null - null --> null
        ('function test() {return(1)}, x = test, y = test', 'x - y', None, True),  # function - function --> ERROR

        # for Evaluator._multiply
        ('', '{"a": 2} * {"a": 10, "b": 3}', {"a": 20, "b": None}, False),  # block * block --> block
        ('', '{"a": 2} * {"b": 3}', {"a": None, "b": None}, False),  # block * block --> block
        ('', '{"a": 2, "b": 3, "c": 4, "d": 5, "e": 6} * {"f": 1}', None, True),  # block * block --> ERROR because max array size is limited as 5.
        ('', '{"a": 2} * ["b", 3]', None, True),  # block * array --> ERROR
        ('', '{"a": 2} * "b"', None, True),  # block * string --> ERROR
        ('', '{"a": 2} * 3', None, True),  # block * int --> ERROR
        ('', '{"a": 2} * 3.5', None, True),  # block * float --> ERROR
        ('', '{"a": 2} * false', None, True),  # block * boolean --> ERROR
        ('', '{"a": 2} * null', None, False),  # block * null --> null
        ('', '[100] * {"b": 3}', None, True),  # array * block --> ERROR
        ('', '[100] * [3]', None, True),  # array * array --> ERROR
        ('', '[100, 3, 1e+2] * "|"', "100|3|100.0", False),  # array * string --> string  (Concatenate elements with delimiter)
        ('', '[100] * 5', [100, 100, 100, 100, 100], False),  # array * int --> array  (for initialization of array)
        ('', '[100] * 6', None, True),  # array * int --> ERROR because max array size is limited as 5.
        ('', '[100, "a"] * 2', [100, "a", 100, "a"], False),  # array * int --> array  (for initialization of array)
        ('', '[100, "a"] * 3', None, True),  # array * int --> ERROR because max array size is limited as 5.
        ('', '[100] * 3.7', [100, 100, 100], False),  # array * float --> array  (for initialization of array)
        ('', '[100, "a"] * 2.7', [100, "a", 100, "a"], False),  # array * float --> array  (for initialization of array)
        ('', '[100, "a"] * 3.7', None, True),  # array * float --> ERROR because max array size is limited as 5.
        ('', '[100] * true', None, True),  # array * boolean --> ERROR
        ('', '[100] * false', None, True),  # array * boolean --> ERROR
        ('', '[100] * null', None, False),  # array * null --> null
        ('', '"|" * {"a", 100}', None, True),  # string * block --> ERROR
        ('', '"|" * [100, "a"]', "100|a", False),  # string * array --> string  (Concatenate elements with delimiter)
        ('', '"3a" * "3a"', None, True),  # string * string --> ERROR
        ('', '"3a" * 2', "3a3a", False),  # string * int --> string
        ('', '"3a" * 2.9', "3a3a", False),  # string * float --> string
        ('', '"3a" * true', None, True),  # string * boolean --> ERROR
        ('', '"3a" * false', None, True),  # string * boolean --> ERROR
        ('', '"3a" * null', None, False),  # string * null --> null
        ('', '3 * {"a": 50}', None, True),  # int * block --> ERROR
        ('', '3 * [50]', [50, 50, 50], False),  # int * array --> array  (for initialization of array)
        ('', '2 * "3a"', "3a3a", False),  # int * string --> string
        ('', '2 * 60', 120, False),  # int * int --> int
        ('', '2 * 60.0', 120.0, False),  # int * float --> float
        ('', '2 * 60.7', 121.4, False),  # int * float --> float
        ('', '2 * true', None, True),  # int * boolean --> ERROR
        ('', '2 * false', None, True),  # int * boolean --> ERROR
        ('', '2 * null', None, False),  # int * null --> null
        ('', '3.8 * {"a": 50}', None, True),  # float * block --> ERROR
        ('', '3.8 * [50]', [50, 50, 50], False),  # float * array --> array  (for initialization of array)
        ('', '2.8 * "3a"', "3a3a", False),  # float * string --> string
        ('', '2.8 * 60', 168.0, False),  # float * int --> float
        ('', '2.8 * 60.7', 169.96, False),  # float * float --> float
        ('', '2.8 * true', None, True),  # float * boolean --> ERROR
        ('', '2.8 * false', None, True),  # float * boolean --> ERROR
        ('', '2.8 * null', None, False),  # float * null --> null
        ('', 'true * {"a": 3}', None, True),  # boolean * block --> ERROR
        ('', 'true * [3, 2]', None, True),  # boolean * array --> ERROR
        ('', 'false * [3, 2]', None, True),  # boolean * array --> ERROR
        ('', 'true * "x1"', None, True),  # boolean * string --> ERROR
        ('', 'false * "x1"', None, True),  # boolean * string --> ERROR
        ('', 'true * 20', None, True),  # boolean * int --> ERROR
        ('', 'false * 20', None, True),  # boolean * int --> ERROR
        ('', 'true * 20.5', None, True),  # boolean * float --> ERROR
        ('', 'false * 20.5', None, True),  # boolean * float --> ERROR
        ('', 'true * true', None, True),  # boolean * boolean --> ERROR
        ('', 'true * false', None, True),  # boolean * boolean --> ERROR
        ('', 'false * true', None, True),  # boolean * boolean --> ERROR
        ('', 'false * false', None, True),  # boolean * boolean --> ERROR
        ('', 'null * {"a": 3}', None, False),  # null * block --> ERROR
        ('', 'null * [3, 2]', None, False),  # null * array --> null
        ('', 'null * "a"', None, False),  # null * string --> null
        ('', 'null * 3', None, False),  # null * int --> null
        ('', 'null * 3.8', None, False),  # null * float --> null
        ('', 'null * true', None, False),  # null * boolean --> null
        ('', 'null * false', None, False),  # null * boolean --> null
        ('', 'null * null', None, False),  # null * null --> null
        ('function test() {return(1)}, x = test, y = test', 'x * y', None, True),  # function * function --> ERROR

        # for Evaluator._divide
        ('', '{"a": 2} / {"a": 10, "b": 3}', {"a": 0.2, "b": None}, False),  # block / block --> block
        ('', '{"a": 2} / {"a": 2, "b": 3, "c": 4, "d": 5, "e": 6, "f": 7}', None, True),  # block / block --> ERROR because max array size is limited as 5.
        ('', '{"a": 2} / {"b": 3}', None, True),  # block / block --> ERROR
        ('', '{"a": 2} / [2]', None, True),  # block / array --> ERROR
        ('', '{"a": 2} / "a"', None, True),  # block / string --> ERROR
        ('', '{"a": 2} / 2', None, True),  # block / int --> ERROR
        ('', '{"a": 2.0} / 2.0', None, True),  # block / float --> ERROR
        ('', '{"a": true} / true', None, True),  # block / boolean --> ERROR
        ('', '{"a": true} / null', None, True),  # block / null --> ERROR
        ('', '[3, 2] / [2]', None, True),  # array / array --> ERROR
        ('', '[3, 2] / ","', None, True),  # array / string --> ERROR
        ('', '[3, 2] / 2', None, True),  # array / int --> ERROR
        ('', '[3, 2] / 2.5', None, True),  # array / float --> ERROR
        ('', '[3, 2] / false', None, True),  # array / boolean --> ERROR
        ('', '[3, 2] / true', None, True),  # array / boolean --> ERROR
        ('', '[3, 2] / null', None, True),  # array / null --> ERROR
        ('', '"a,b" / {"a": ","}', None, True),  # string / block --> ERROR
        ('', '"a,b" / [","]', None, True),  # string / array --> ERROR
        ('', '"a,b" / ","', ["a", "b"], False),  # string / string --> array
        ('', '"a,b," / ","', ["a", "b", ""], False),  # string / string --> array
        ('', '"a,b,c,d,e" / ","', ["a", "b", "c", "d", "e"], False),  # string / string --> array
        ('', '"a,b,c,d,e,f" / ","', None, True),  # string / string --> ERROR because max array size is limited as 5.
        ('', '"a,b" / ""', ["a", ",", "b"], False),  # string / string --> array
        ('', '"xxwwwwyy" / "ww"', ['xx', '', 'yy'], False),  # string / string --> array
        ('', '"xxwwwyy" / "ww"', ['xx', 'wyy'], False),  # string / string --> array
        ('', '"abcde" / ""', ["a", "b", "c", "d", "e"], False),  # string / string --> array
        ('', '"abcdef" / ""', None, True),  # string / string --> ERROR because max array size is limited as 5.
        ('', '"a,b" / 2', None, True),  # string / int --> ERROR
        ('', '"a,b" / 2.5', None, True),  # string / float --> ERROR
        ('', '"a,b" / true', None, True),  # string / boolean --> ERROR
        ('', '"a,b" / false', None, True),  # string / boolean --> ERROR
        ('', '"a,b" / null', None, True),  # string / null --> ERROR
        ('', '20 / {"a": 2}', None, True),  # int / block --> ERROR
        ('', '20 / [2, 5]', None, True),  # int / array --> ERROR
        ('', '20 / "0"', None, True),  # int / string --> ERROR
        ('', '20 / 5', 4, False),  # int / int --> int or float or ERROR
        ('', '20 / 8', 2.5, False),  # int / int --> int or float or ERROR
        ('', '20 / 0', None, True),  # int / 0 --> ERROR
        ('', '20 / 0.0', None, True),  # int / 0.0 --> ERROR
        ('', '20 / 2.5', 8, False),  # int / float --> int or float or ERROR
        ('', '20 / 8', 2.5, False),  # int / float --> int or float or ERROR
        ('', '20 / null', None, True),  # int / null --> ERROR
        ('', '3.5 / {"a": 2.5}', None, True),  # float / block --> ERROR
        ('', '3.5 / [1, 2]', None, True),  # float / array --> ERROR
        ('', '3.5 / "."', None, True),  # float / string --> ERROR
        ('', '3.5 / "0.5"', None, True),  # float / string --> ERROR
        ('', '3.5 / 5', 0.7, False),  # float / int --> int or float or ERROR
        ('', '3.0 / 3', 1, False),  # float / int --> int or float or ERROR
        ('', '3.0 / 0', None, True),  # float / 0 --> ERROR
        ('', '3.5 / 0.5', 7, False),  # float / float --> int or float or ERROR
        ('', '0.49 / 0.7', 0.7, False),  # float / float --> int or float or ERROR
        ('', '0.49 / 0.0', None, True),  # float / 0.0 --> ERROR
        ('', '3.5 / true', None, True),  # float / boolean --> ERROR
        ('', '3.5 / false', None, True),  # float / boolean --> ERROR
        ('', '3.5 / null', None, True),  # float / null --> ERROR
        ('', 'true / {"true": true}', None, True),  # boolean / block --> ERROR
        ('', 'true / [true]', None, True),  # boolean / array --> ERROR
        ('', 'false / [true]', None, True),  # boolean / array --> ERROR
        ('', 'true / [false]', None, True),  # boolean / array --> ERROR
        ('', 'false / [false]', None, True),  # boolean / array --> ERROR
        ('', 'true / "true"', None, True),  # boolean / string --> ERROR
        ('', 'false / "false"', None, True),  # boolean / string --> ERROR
        ('', 'true / 1', None, True),  # boolean / int --> ERROR
        ('', 'false / 1', None, True),  # boolean / int --> ERROR
        ('', 'true / 1.5', None, True),  # boolean / float --> ERROR
        ('', 'false / 1.5', None, True),  # boolean / float --> ERROR
        ('', 'true / true', None, True),  # boolean / boolean --> ERROR
        ('', 'false / true', None, True),  # boolean / boolean --> ERROR
        ('', 'true / null', None, True),  # boolean / null --> ERROR
        ('', 'false / null', None, True),  # boolean / null --> ERROR
        ('', 'null / {"a": null}', None, False),  # null / block --> null
        ('', 'null / [null]', None, False),  # null / array --> null
        ('', 'null / [1, 2]', None, False),  # null / array --> null
        ('', 'null / "a"', None, False),  # null / string --> null
        ('', 'null / 2', None, False),  # null / int --> null
        ('', 'null / 0', None, True),  # null / int --> ERROR
        ('', 'null / 8.8', None, False),  # null / float --> null
        ('', 'null / 0.0', None, True),  # null / float --> ERROR
        ('', 'null / false', None, False),  # null / boolean --> null
        ('', 'null / null', None, True),  # null / null --> ERROR
        ('function test() {return(1)}, x = test, y = test', 'x / y', None, True),  # function / function --> ERROR

        # for Evaluator._modulo
        ('', '{"a": 20} % {"a": 6, "b": 3}', {"a": 2, "b": None}, False),  # block % block --> block
        ('', '{"a": 20} % {"a": 2, "b": 3, "c": 4, "d": 5, "e": 6, "f": 7}', None, True),  # block % block --> ERROR because max array size is limited as 5.
        ('', '{a: 20} % {a: 6, b: 3}', {"a": 2, "b": None}, False),  # block % block --> block
        ('', '{"a": 2} % {"b": 3}', None, True),  # block % block --> ERROR
        ('', '{"a": 2} % [2]', None, True),  # block % array --> ERROR
        ('', '{"a": 2} % "a"', None, True),  # block % string --> ERROR
        ('', '{"a": 2} % 2', None, True),  # block % int --> ERROR
        ('', '{"a": 2} % 1.5', None, True),  # block % float --> ERROR
        ('', '{"a": 2} % true', None, True),  # block % boolean --> ERROR
        ('', '{"a": 2} % null', None, True),  # block % null --> ERROR
        ('', '[2, 3] % {"a": 2}', None, True),  # array % block --> ERROR
        ('', '[2, 3] % [2, 3]', None, True),  # array % array --> ERROR
        ('', '[2, 3] % ","', None, True),  # array % string --> ERROR
        ('', '[2, 3] % 2', None, True),  # array % int --> ERROR
        ('', '[2, 3] % 2.5', None, True),  # array % float --> ERROR
        ('', '[2, 3] % true', None, True),  # array % boolean --> ERROR
        ('', '[2, 3] % null', None, True),  # array % null --> ERROR
        ('', '"a, 2" % {"a", 2}', None, True),  # string % block --> ERROR
        ('', '"2, 3" % [2]', None, True),  # string % array --> ERROR
        ('', '"2, 3" % ","', None, True),  # string % string --> ERROR
        ('', '"2, 3" % 2', None, True),  # string % int --> ERROR
        ('', '"2, 3" % 2.0', None, True),  # string % float --> ERROR
        ('', '"2, 3" % true', None, True),  # string % boolean --> ERROR
        ('', '"2, 3" % null', None, True),  # string % null --> ERROR
        ('', '10 % {"a": 2}', None, True),  # int % block --> ERROR
        ('', '10 % [3]', None, True),  # int % array --> ERROR
        ('', '10 % "3"', None, True),  # int % string --> ERROR
        ('', '10 % 3', 1, False),  # int % int --> int or float or ERROR, 10 = 3 x 3 + 1
        ('', '-10 % 3', 2, False),  # int % int --> int or float or ERROR, -10 = 3 x (-4) + 2 (NG: -10 = 3 x (-3) - 1)
        ('', '10 % -3', -2, False),  # int % int --> int or float or ERROR, 10 = -3 x (-4) - 2 (NG: 10 = -3 x (-3) + 1)
        ('', '-10 % -3', -1, False),  # int % int --> int or float or ERROR, -10 = -3 x 3 - 1 (NG: -10 = -3 x 4 + 2)
        ('', '10 % 0', None, True),  # int % 0 --> ERROR
        ('', '10 % 3.0', 1, False),  # int % float --> int or float or ERROR
        ('', '10 % 1.7', 1.5, False),  # int % float --> int or float or ERROR
        ('', '10 % 0.0', None, True),  # int % 0.0 --> ERROR
        ('', '10 % true', None, True),  # int % boolean --> ERROR
        ('', '10 % null', None, True),  # int % null --> ERROR
        ('', '10.5 % {"a": 3.0}', None, True),  # float % block --> ERROR
        ('', '10.5 % [3.0]', None, True),  # float % array --> ERROR
        ('', '10.5 % "a"', None, True),  # float % string --> ERROR
        ('', '10.5 % 3', 1.5, False),  # float % int --> int or float or ERROR
        ('', '10.5 % 0', None, True),  # float % 0 --> ERROR
        ('', '17.5 % 1.5', 1, False),  # float % float --> int or float or ERROR
        ('', '10.5 % 1.7', 0.3, False),  # float % float --> int or float or ERROR
        ('', '10.5 % 0.0', None, True),  # float % 0.0 --> ERROR
        ('', '10.5 % true', None, True),  # float % boolean --> ERROR
        ('', '10.5 % null', None, True),  # float % null --> ERROR
        ('', 'true % {"true": true}', None, True),  # boolean % block --> ERROR
        ('', 'true % [true]', None, True),  # boolean % array --> ERROR
        ('', 'true % "true"', None, True),  # boolean % string --> ERROR
        ('', 'true % 3', None, True),  # boolean % int --> ERROR
        ('', 'true % 1.7', None, True),  # boolean % float --> ERROR
        ('', 'true % true', None, True),  # boolean % boolean --> ERROR
        ('', 'true % null', None, True),  # boolean % null --> ERROR
        ('', 'null % {"a": null}', None, False),  # null % block --> null
        ('', 'null % [null]', None, False),  # null % array --> null
        ('', 'null % "null"', None, False),  # null % string --> null
        ('', 'null % 3', None, False),  # null % int --> null
        ('', 'null % 0', None, True),  # null % int --> ERROR
        ('', 'null % 1.7', None, False),  # null % float --> null
        ('', 'null % 0.0', None, True),  # null % float --> ERROR
        ('', 'null % true', None, False),  # null % boolean --> null
        ('', 'null % null', None, True),  # null % null --> ERROR
        ('function test() {return(1)}, x = test, y = test', 'x % y', None, True),  # function % function --> ERROR

        # for Evaluator._logical_invert
        ('a = {}', 'not a', False, False),  # not block --> boolean (using truthy) (always false)
        ('a = {"a": 3}', 'not a', False, False),  # not block --> boolean (using truthy) (always false)
        ('', 'not {"a": 3}', None, True),  # not {"a": 3} is confused as function "not". --> ERROR
        ('a = []', 'not a', False, False),  # not array --> boolean (using truthy) (always false)
        ('a = [3]', 'not a', False, False),  # not array --> boolean (using truthy) (always false)
        ('', 'not [3]', None, True),  # not [3] is confused as variable "not". --> ERROR
        ('', 'not ""', False, False),  # not string --> boolean (using truthy) (always false)
        ('', 'not 0', True, False),  # not int --> boolean (using truthy)
        ('', 'not 10', False, False),  # not int --> boolean (using truthy)
        ('', 'not 0.0', True, False),  # not float --> boolean (using truthy)
        ('', 'not 5.0', False, False),  # not float --> boolean (using truthy)
        ('', 'not true', False, False),  # not boolean --> boolean
        ('', 'not false', True, False),  # not boolean --> boolean
        ('', 'not null', True, False),  # not null --> boolean (using truthy) (always true)
        ('', 'false not false', None, True),  # "not" is binary expression in grammar.
        ('function test() {return(1)}, x = test', 'not x', False, False),  # not function --> boolean (using truthy)

        # for Evaluator._logical_and
        ('', 'false and false', False, False),
        ('', 'false and true', False, False),
        ('', 'true and false', False, False),
        ('', 'true and true', True, False),
        ('', '{} and false', False, False),  # using truthy
        ('', '{} and true', True, False),  # using truthy
        ('', '{"a": 3} and false', False, False),  # using truthy
        ('', '{"a": 3} and true', True, False),  # using truthy
        ('', '[] and false', False, False),  # using truthy
        ('', '[] and true', True, False),  # using truthy
        ('', '[3] and false', False, False),  # using truthy
        ('', '[3] and true', True, False),  # using truthy
        ('', '"a" and false', False, False),  # using truthy
        ('', '"a" and true', True, False),  # using truthy
        ('', '10 and false', False, False),  # using truthy
        ('', '10 and true', True, False),  # using truthy
        ('', '0 and false', False, False),  # using truthy, 0 is equivalent to false.
        ('', '0 and true', False, False),  # using truthy, 0 is equivalent to false.
        ('', '10.5 and false', False, False),  # using truthy
        ('', '10.5 and true', True, False),  # using truthy
        ('', '0.0 and false', False, False),  # using truthy, 0 is equivalent to false.
        ('', '0.0 and true', False, False),  # using truthy, 0 is equivalent to false.
        ('', 'null and false', False, False),  # using truthy, null is equivalent to false.
        ('', 'null and true', False, False),  # using truthy, null is equivalent to false.
        ('', 'false and {}', False, False),  # using truthy
        ('', 'true and {}', True, False),  # using truthy
        ('', 'false and {"a": 3}', False, False),  # using truthy
        ('', 'true and {"a": 3}', True, False),  # using truthy
        ('', 'false and []', False, False),  # using truthy
        ('', 'true and []', True, False),  # using truthy
        ('', 'false and [3]', False, False),  # using truthy
        ('', 'true and [3]', True, False),  # using truthy
        ('', 'false and "a"', False, False),  # using truthy
        ('', 'true and "a"', True, False),  # using truthy
        ('', 'false and 10', False, False),  # using truthy
        ('', 'true and 10', True, False),  # using truthy
        ('', 'false and 0', False, False),  # using truthy, 0 is equivalent to false.
        ('', 'true and 0', False, False),  # using truthy, 0 is equivalent to false.
        ('', 'false and 10.5', False, False),  # using truthy
        ('', 'true and 10.5', True, False),  # using truthy
        ('', 'false and 0.0', False, False),  # using truthy, 0 is equivalent to false.
        ('', 'true and 0.0', False, False),  # using truthy, 0 is equivalent to false.
        ('', 'false and null', False, False),  # using truthy, null is equivalent to false.
        ('', 'true and null', False, False),  # using truthy, null is equivalent to false.
        ('y = 0; function y10() {y = 10, return(true)}, z = 0; if (false or y10()) {z = 20}', '[y, z]', [10, 20], False),
        ('y = 0; function y10() {y = 10, return(true)}, z = 0; if (true or y10()) {z = 20}', '[y, z]', [0, 20], False),  # Short-circuit evaluation
        ('y = 0; function y10() {y = 10, return(true)}, z = 0; if (false and y10()) {z = 20}', '[y, z]', [0, 0], False),  # Short-circuit evaluation
        ('y = 0; function y10() {y = 10, return(true)}, z = 0; if (true and y10()) {z = 20}', '[y, z]', [10, 20], False),
        ('function test() {return(1)}, x = test, y = test', 'x and y', True, False),  # using truthy

        # for Evaluator._logical_or
        ('', 'false or false', False, False),
        ('', 'false or true', True, False),
        ('', 'true or false', True, False),
        ('', 'true or true', True, False),
        ('', '{} or false', True, False),  # using truthy
        ('', '{} or true', True, False),  # using truthy
        ('', '{"a": 3} or false', True, False),  # using truthy
        ('', '{"a": 3} or true', True, False),  # using truthy
        ('', '[] or false', True, False),  # using truthy
        ('', '[] or true', True, False),  # using truthy
        ('', '[3] or false', True, False),  # using truthy
        ('', '[3] or true', True, False),  # using truthy
        ('', '"a" or false', True, False),  # using truthy
        ('', '"a" or true', True, False),  # using truthy
        ('', '10 or false', True, False),  # using truthy
        ('', '10 or true', True, False),  # using truthy
        ('', '0 or false', False, False),  # using truthy, 0 is equivalent to false.
        ('', '0 or true', True, False),  # using truthy, 0 is equivalent to false.
        ('', '10.5 or false', True, False),  # using truthy
        ('', '10.5 or true', True, False),  # using truthy
        ('', '0.0 or false', False, False),  # using truthy, 0 is equivalent to false.
        ('', '0.0 or true', True, False),  # using truthy, 0 is equivalent to false.
        ('', 'null or false', False, False),  # using truthy, null is equivalent to false.
        ('', 'null or true', True, False),  # using truthy, null is equivalent to false.
        ('', 'false or {}', True, False),  # using truthy
        ('', 'true or {}', True, False),  # using truthy
        ('', 'false or {"a": 3}', True, False),  # using truthy
        ('', 'true or {"a": 3}', True, False),  # using truthy
        ('', 'false or []', True, False),  # using truthy
        ('', 'true or []', True, False),  # using truthy
        ('', 'false or [3]', True, False),  # using truthy
        ('', 'true or [3]', True, False),  # using truthy
        ('', 'false or "a"', True, False),  # using truthy
        ('', 'true or "a"', True, False),  # using truthy
        ('', 'false or 10', True, False),  # using truthy
        ('', 'true or 10', True, False),  # using truthy
        ('', 'false or 0', False, False),  # using truthy, 0 is equivalent to false.
        ('', 'true or 0', True, False),  # using truthy, 0 is equivalent to false.
        ('', 'false or 10.5', True, False),  # using truthy
        ('', 'true or 10.5', True, False),  # using truthy
        ('', 'false or 0.0', False, False),  # using truthy, 0 is equivalent to false.
        ('', 'true or 0.0', True, False),  # using truthy, 0 is equivalent to false.
        ('', 'false or null', False, False),  # using truthy, null is equivalent to false.
        ('', 'true or null', True, False),  # using truthy, null is equivalent to false.
        ('function test() {return(1)}, x = test, y = test', 'x or y', True, False),  # using truthy

        # for Evaluator._is_equal
        ('', '{"a": 1, "b": 2} == {"b": 2, "a": 1}', True, False),
        ('', '{"a": 1, "b": 2} == {"a": 1}', False, False),
        ('', '[1, 2] == [1, 2]', True, False),
        ('', '[1, 2] == [2, 1]', False, False),
        ('', '"a" == "a"', True, False),
        ('', '"a" == "b"', False, False),
        ('', '100 == 100', True, False),
        ('', '100 == 200', False, False),
        ('', '1.5 == 1.5', True, False),
        ('', '1.5 == 1.4', False, False),
        ('', 'true == true', True, False),
        ('', 'true == false', False, False),
        ('', 'null == null', True, False),
        ('a = null', 'null == a', True, False),
        ('', '{"a": 1, "b": 2} == ["a", "b"]', False, False),
        ('', '{"a": 1, "b": 2} == [1, 2]', False, False),
        ('', '{"a": 1, "b": 2} == true', True, False),  # using truthy
        ('', '[1, 2] == true', True, False),  # using truthy
        ('', '[1, 2] == 1', False, False),  # using truthy
        ('', '"a" == true', True, False),  # using truthy
        ('', 'true == "a"', True, False),  # using truthy
        ('', '"a" == 1', False, False),  # using truthy
        ('', '100 == true', True, False),  # using truthy
        ('', 'true == 100', True, False),  # using truthy
        ('', '0 == false', True, False),  # using truthy
        ('', 'false == 0', True, False),  # using truthy
        ('', 'null == false', True, False),  # using truthy
        ('', 'false == null', True, False),  # using truthy
        ('function test() {return(1)}, x = test, y = test', 'x == y', True, False),

        # for Evaluator._is_not_equal
        ('', '{"a": 1, "b": 2} != {"b": 2, "a": 1}', False, False),
        ('', '{"a": 1, "b": 2} != {"a": 1}', True, False),
        ('', '[1, 2] != [1, 2]', False, False),
        ('', '[1, 2] != [2, 1]', True, False),
        ('', '"a" != "a"', False, False),
        ('', '"a" != "b"', True, False),
        ('', '100 != 100', False, False),
        ('', '100 != 200', True, False),
        ('', '1.5 != 1.5', False, False),
        ('', '1.5 != 1.4', True, False),
        ('', 'true != true', False, False),
        ('', 'true != false', True, False),
        ('', 'null != null', False, False),
        ('a = null', 'null != a', False, False),
        ('', '{"a": 1, "b": 2} != ["a", "b"]', True, False),
        ('', '{"a": 1, "b": 2} != [1, 2]', True, False),
        ('', '{"a": 1, "b": 2} != false', True, False),  # using truthy
        ('', '{"a": 1, "b": 2} != true', False, False),  # using truthy
        ('', '[1, 2] != true', False, False),  # using truthy
        ('', '[1, 2] != 1', True, False),  # using truthy
        ('', '"a" != true', False, False),  # using truthy
        ('', 'true != "a"', False, False),  # using truthy
        ('', '"a" != 1', True, False),  # using truthy
        ('', '100 != true', False, False),  # using truthy
        ('', 'true != 100', False, False),  # using truthy
        ('', '0 != false', False, False),  # using truthy
        ('', 'false != 0', False, False),  # using truthy
        ('', 'null != false', False, False),  # using truthy
        ('', 'false != null', False, False),  # using truthy
        ('function test() {return(1)}, x = test, y = test', 'x != y', False, False),

        # for Evaluator._is_less_than
        ('', '3 < 5', True, False),
        ('', '3 < 3', False, False),
        ('', '-3 < 2', True, False),
        ('', '3 < -5', False, False),
        ('', '2.3 < 2.4', True, False),
        ('', '1e4 < 1e3', False, False),
        ('', '1e-4 < 1e-3', True, False),
        ('', '{} < 5', None, True),
        ('', '[] < 5', None, True),
        ('', '"" < 5', None, True),
        ('', '"0" < 5', None, True),
        ('', 'false < 5', None, True),
        ('', 'null < 5', None, True),
        ('function test() {return(1)}, x = test, y = test', 'x < y', None, True),

        # for Evaluator._is_less_than_or_equal
        ('', '3 <= 5', True, False),
        ('', '3 <= 3', True, False),
        ('', '-3 <= 2', True, False),
        ('', '3 <= -5', False, False),
        ('', '2.3 <= 2.3', True, False),
        ('', '2.3 <= 2.4', True, False),
        ('', '-1e+4 <= -1e+4', True, False),
        ('', '1e4 <= 1e3', False, False),
        ('', '1e-4 <= 1e-3', True, False),
        ('', '{} <= 5', None, True),
        ('', '[] <= 5', None, True),
        ('', '"" <= 5', None, True),
        ('', '"0" <= 5', None, True),
        ('', 'false <= 5', None, True),
        ('', 'null <= 5', None, True),
        ('function test() {return(1)}, x = test, y = test', 'x <= y', None, True),

        # for Evaluator._is_greater_than
        ('', '3 > 1', True, False),
        ('', '3 > 5', False, False),
        ('', '3 > 3', False, False),
        ('', '-3 > 2', False, False),
        ('', '3 > -5', True, False),
        ('', '2.3 < 2.4', True, False),
        ('', '1e4 > 1e3', True, False),
        ('', '1e-4 > 1e-3', False, False),
        ('', '{} > 5', None, True),
        ('', '[] > 5', None, True),
        ('', '"" > 5', None, True),
        ('', '"0" > 5', None, True),
        ('', 'false > 5', None, True),
        ('', 'null > 5', None, True),
        ('function test() {return(1)}, x = test, y = test', 'x > y', None, True),

        # for Evaluator._is_greater_than_or_equal
        ('', '3 >= 1', True, False),
        ('', '3 >= 5', False, False),
        ('', '3 >= 3', True, False),
        ('', '-3 >= 2', False, False),
        ('', '3 >= -5', True, False),
        ('', '2.3 < 2.4', True, False),
        ('', '1e3 >= 1e3', True, False),
        ('', '1e4 >= 1e3', True, False),
        ('', '1e-4 >= 1e-3', False, False),
        ('', '{} >= 5', None, True),
        ('', '[] >= 5', None, True),
        ('', '"" >= 5', None, True),
        ('', '"0" >= 5', None, True),
        ('', 'false >= 5', None, True),
        ('', 'null >= 5', None, True),
        ('function test() {return(1)}, x = test, y = test', 'x >= y', None, True),

        # for Evaluator._is_contained
        ('', '"b" in ["a", "b", "c"]', True, False),
        ('', '"d" in ["a", "b", "c"]', False, False),
        ('', '1 in [1, 2, "c"]', True, False),
        ('', '3 in [1, 2, "c"]', False, False),
        ('', 'true in [1, true, "c"]', True, False),
        ('', '[2, 3] in [1, [2,3], "c"]', True, False),
        ('', '{"a": 2, "b": false} in [1, {"b":false, "a":2}, "c"]', True, False),
        ('', '"x" in {"x": 3, "y": 10}', True, False),
        ('', '"z" in {"x": 3, "y": 10}', False, False),
        ('', '{"x": 3} in {"x": 3, "y": 10}', True, False),
        ('', '{"x": 2} in {"x": 3, "y": 10}', False, False),
        ('', '{"x": 3, "y": 10} in {"x": 3, "y": 10}', True, False),
        ('', '{"x": 3, "y": 1} in {"x": 3, "y": 10}', False, False),
        ('', '"dog" in "cat&dog&bird"', True, False),
        ('', '"dogs" in "cat&dog&bird"', False, False),
        ('', '"23" in "123"', True, False),
        ('', '23 in "123"', None, True),  # int in str --> ERROR
        ('', '1 in 5', None, True),  # int in int --> ERROR
        ('', '1.5 in 4.5', None, True),  # float in float --> ERROR
        ('', '[1, 2, "c"] in 1', None, True),  # array in int --> ERROR
        ('', 'true in true', None, True),  # boolean in boolean --> ERROR
        ('', 'null in null', None, True),  # null in null --> ERROR
        ('function test() {return(1)}, x = test, y = test', 'x in y', None, True),
        ('function test() {return(1)}, x = test, y = test', '1 in y', None, True),
        ('function test() {return(1)}, x = test, y = test', 'x in [y]', True, False),
        ('{"foo":3, "bar": 2}', '"foo" in .', True, False),
        ('{"foo":3, "bar": 2}', '"baz" in .', False, False),
    )
)

def test_binary(prepared_expression, target_expression, expected_value,
                is_error):
    """Unit-test for binary expression in Evaluator

    Unit-test for "+", "*", "and", "<=", ...

    Args:
        prepared_expression (str or dict): When the another expression,
                                 such as assignment, is needed before
                                 test, it is given here.
                                 Maybe this is '' mostly.
                                 str: Statements as VivJson
                                 dict: Value as Python's value
        target_expression (str): "return" method's argument as targeted
                                 expression.
        expected_value (Any): The expected value as Python's value
        is_error (bool): True if occurring error is expected,
                         False otherwise.
    """
    config = Config(max_array_size=5)

    text = f'return({target_expression})'
    value, error_message = Viv.run(prepared_expression, text, config)
    assert (error_message != '') == is_error

    if isinstance(value, float):
        value = round(value, 10)
    assert value == expected_value

    if error_message == '' \
            and not isinstance(expected_value, bool) \
            and isinstance(expected_value, (int, float)):
        assert type(value) == type(expected_value)  # pylint: disable=C0123



"""Error for VivJson

- Error.make_lex_error: Report error information and return LexError.
- Error.make_parse_error: Report error information and return
                          ParseError.
- Error.make_evaluate_error: Report error information and return
                             EvaluateError.

Refer to:
- "Crafting Interpreters"
  https://craftinginterpreters.com/
Note that this code is made from scratch. The source code
of the above WEB site is not used.

Environment:
- Python 3.9 or later

License:
Copyright 2025 benesult

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

__author__ = 'Fumiaki Motegi <motegi@benesult.com>'
__date__ = '2025-03-22'

from functools import singledispatchmethod
import sys
from .config import Config
from .statement import Literal, Identifier, Keyword, Blank, \
                       Array, Block, Binary, Parameter, \
                       Callee, Call, CalleeRegistry, Loop, \
                       Get, Set, Remove, \
                       Return, Injection, Value
from .tokens import Token

class Error:
    """Error class"""
    TAG = 'Viv'

    class LexError(RuntimeError):
        """Lex-error class of Runtime error"""

    class ParseError(RuntimeError):
        """Parse-error class of Runtime error"""

    class EvaluateError(RuntimeError):
        """Evaluate-error class of Runtime error"""

    @staticmethod
    def make_lex_error(message, medium, line, column, config=None):
        """Make lex-error.

        Report error information and return LexError that is Exception.
        This method does not raise Exception. So caller should raise
        Exception if it is needed.

        Args:
            message (str): Error message
            medium (str or NoneType): The object that has source code.
            line (int): Line number of location where error occurs.
            column (int): Column number of location where error occurs.
            config (Config or NoneType, optional): Configuration if
                                                   needed.

        Returns:
            LexError: Exception for raising.
        """
        report = Error.report('Lexer', message,
                      Error._get_location_text(medium, line, column), config)
        sys.tracebacklimit = 0
        return Error.LexError(report)

    @staticmethod
    def make_parse_error(message, token=None, config=None):
        """Make parse-error.

        Report error information and return ParseError that is
        Exception.
        This method does not raise Exception. So caller should raise
        Exception if it is needed.

        Args:
            message (str): Error message
            token (Token or NoneType, optional): Token.
                                                 None is given when
                                                 cause is not Token.
            config (Config or NoneType, optional): Configuration if
                                                   needed.

        Returns:
            ParseError: Exception for raising.
        """
        report = Error.report('Parser', message, token, config)
        sys.tracebacklimit = 0
        return Error.ParseError(report)

    @staticmethod
    def make_evaluate_error(message, statement=None, config=None,
                            eliminate_tag=False):
        """Make evaluate-error.

        Report error information and return EvaluateError that is
        Exception.
        This method does not raise Exception. So caller should raise
        Exception if it is needed.

        Args:
            message (str): Error message
            statement (Statement or NoneType, optional): Statement
                    where error occurs. It is used to get location.
                    None is given when cause is not Statement.
            config (Config or NoneType, optional): Configuration if
                                                   needed.
            eliminate_tag (bool, optional): True if tag is not needed,
                                            False otherwise.
                                            The default is False

        Returns:
            EvaluateError: Exception for raising.
        """
        location = Error._get_statement_location(statement)
        report = Error.report('Evaluator', message, location, config,
                               eliminate_tag)
        sys.tracebacklimit = 0
        return Error.EvaluateError(report)

    @staticmethod
    def report(tag, message, location=None, config=None,
               eliminate_tag=False):
        """Report an error message.

        Print an error message into stderr according as configuration.

        Args:
            tag (str): Tag, such as "Lexer", "Parser", or "Evaluator".
            message (str): Main message of Error
            location (Token or str or NoneType, optional): The location
                                                           of error
            config (Config or NoneType, optional): Configuration if
                                                   needed.
            eliminate_tag (bool, optional): True if tag is not needed,
                                            False otherwise.
                                            The default is False

        Returns:
            str: Error message
        """
        tag_detail = ''
        if Config.get_enable_tag_detail_alt(config):
            tag_detail = f':{tag}'
        text = message if eliminate_tag \
                    else  f'[{Error.TAG}{tag_detail}] Error: {message}'
        if location is not None:
            if isinstance(location, Token):
                text = f'{text} for {location.to_string()}'
            elif isinstance(location, str):
                text = f'{text} in {location}'
        if Config.get_enable_stderr_alt(config):
            print(text, file=sys.stderr)
        return text

    @staticmethod
    def _get_location_text(medium, line: int, column: int) -> str:
        medium = f'{medium}, ' if isinstance(medium, str) else ''
        return f'({medium}line: {line}, column: {column})'

    @staticmethod
    def _get_token_location(token: Token):
        """Get Token's location in source code."""
        location = None
        if token is not None and token.line_number is not None \
                and token.column_number is not None:
            location = Error._get_location_text(token.medium,
                                                token.line_number,
                                                token.column_number)
        return location

    @singledispatchmethod
    @staticmethod
    def _get_statement_location(_):
        """Get Statement's location in source code."""
        return None

    @staticmethod
    @_get_statement_location.register
    def _(literal: Literal):
        return Error._get_token_location(literal.token)

    @staticmethod
    @_get_statement_location.register
    def _(identifier: Identifier):
        return Error._get_token_location(identifier.name)

    @staticmethod
    @_get_statement_location.register
    def _(keyword: Keyword):
        return Error._get_token_location(keyword.token)

    @staticmethod
    @_get_statement_location.register
    def _(blank: Blank):
        return Error._get_token_location(blank.token)

    @staticmethod
    @_get_statement_location.register
    def _(array: Array):
        if len(array.values) == 0:
            return None
        return Error._get_statement_location(array.values[0])

    @staticmethod
    @_get_statement_location.register
    def _(block: Block):
        if len(block.values) == 0:
            return None
        return Error._get_statement_location(block.values[0])

    @staticmethod
    @_get_statement_location.register
    def _(binary: Binary):
        return Error._get_statement_location(binary.left)

    @staticmethod
    @_get_statement_location.register
    def _(parameter: Parameter):
        if parameter.modifier is not None:
            return Error._get_token_location(parameter.modifier)
        return Error._get_statement_location(parameter.name)

    @staticmethod
    @_get_statement_location.register
    def _(callee: Callee):
        return Error._get_statement_location(callee.name)

    @staticmethod
    @_get_statement_location.register
    def _(callee_registry: CalleeRegistry):
        return Error._get_statement_location(callee_registry.callee)

    @staticmethod
    @_get_statement_location.register
    def _(call: Call):
        return Error._get_statement_location(call.name)

    @staticmethod
    @_get_statement_location.register
    def _(loop: Loop):
        return Error._get_statement_location(loop.call)

    @staticmethod
    @_get_statement_location.register
    def _(get: Get):
        if len(get.members) == 0:
            return None
        return Error._get_statement_location(get.members[0])

    @staticmethod
    @_get_statement_location.register
    def _(s: Set):
        if len(s.members) == 0:
            return None
        return Error._get_statement_location(s.members[0])

    @staticmethod
    @_get_statement_location.register
    def _(remove: Remove):
        return Error._get_token_location(remove.token)

    @staticmethod
    @_get_statement_location.register
    def _(r: Return):
        return Error._get_token_location(r.token)

    @staticmethod
    @_get_statement_location.register
    def _(injection: Injection):
        return injection.location

    @staticmethod
    @_get_statement_location.register
    def _(value: Value):
        return value.location



"""Unit-test for function in Evaluator

Unit-test for function call and its definition

Run `python3 -m pytest tests/test_evaluator_function.py` in
parent directory.

Environment:
- Python 3.9 or later
- pytest 8.3 or later

License:
Copyright 2025 benesult

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

__author__ = 'Fumiaki Motegi <motegi@benesult.com>'
__date__ = '2025-03-30'

# pylint: disable=C0301, E0401, R0801

import pytest
from vivjson.viv import Viv
from vivjson.config import Config

@pytest.mark.parametrize(
    'prepared_expression, target_expression, expected_value, is_error', (
        ('function x2(a) {return(a * 2)}', 'x2(1.5)', 3.0, False),
        ('function x2(a) {return(a*2)}', 'x2(3)', 6, False),
        ('x = test(1), function test(a) {return(a*100)}', 'x', None, True),  # Error because certain function is executed before its definition.
        ('function test(a) {return(a*100)}, x = [test(1)], function test(a) {return(a+10)}, x += test(1)', 'x', [100, 11], False),
        ('function test() {for (i = 0; i < 10; i += 1) { if (i > 5) {return(i)}}}', 'test()', 6, False),
        ('function test() {i = 0, for (; i < 10; i += 1) { if (i > 5) {return(i)}}}', 'test()', 6, False),
        ('function test() {i = 0, for (; i < 10; i += 1) { if (i > 5) {break}}, return(i)}', 'test()', 6, False),
        ('function test() {i = 0, for (; i < 10; i += 1) { if (i > 5) {break}}}', 'test()', {"i": 6}, False),
        ('function test(a) {for (i = 0; i < 10; i += 1) {a /= 2}, return(a)}', 'test(100)', 0.09765625, False),
        ('function test(a) {for (i = 0; i < 10; i += 1) { if (i >= 5) {continue}, a /= 2}, return(a)}', 'test(100)', 3.125, False),
        ('function test(a) {for (i = 0; i < 10; i += 1) { if (i >= 5) {break}, a /= 2}, return(a)}', 'test(100)', 3.125, False),
        ('function test(a) {i = 0, for (; i < 10; i += 1) { if (i >= 5) {continue}, a /= 2}, b = a}', 'test(100)', {"i": 10, "b": 3.125}, False),
        ('function test(a) {i = 0, for (; i < 10; i += 1) { if (i >= 5) {break}, a /= 2}, b = a}', 'test(100)', {"i": 5, "b": 3.125}, False),
        ('function test(a) {i = 0, for (; i < 10; i += 1) { if (i >= 5) {return("foo")}, a /= 2}, b = a}', 'test(100)', "foo", False),
        ('function test() {sum = 0 for (i = 0; i < 10; i += 1) {sum += i}}, a = test()', 'a', {'sum': 45}, False),
        ('function test() {sum = 0 for (i = 0; i < 10; i += 1) {sum += i} := sum}, a = test()', 'a', 45, False),
        ('function test() {sum = 0 for (i = 0; i < 10; i += 1) {sum += i} := sum sum += 100}, a = test()', 'a', 45, False),
        ('function test() {sum = 0 for (i = 0; i < 10; i += 1) {sum += i} := sum sum += 100 return(sum)}, a = test()', 'a', 145, False),
        ('function test() {a = 10}, x = test()', 'x', {"a": 10}, False),
        ('function test() {a = 10, a = 20}, x = test()', 'x', {"a": 20}, False),
        ('function test() {a = 10, a += 20}, x = test()', 'x', {"a": 30}, False),
        ('function test() {a = 10, break, a += 20}, x = test()', 'x', None, True),  # It is invalid that there is "break" just under the function.
        ('function test() {a = 10, continue, a += 20}, x = test()', 'x', None, True),  # It is invalid that there is "continue" just under the function.
        ('function test() {a = 10, return, a += 20}, x = test()', 'x', {"a": 10}, False),
        ('function add(a, b) {return(a + b)} x = add(3, 2, 1)', 'x', 5, False),
        ('function add_dirty(a, b) {return(a + b + _[2])}, x = add_dirty(3, 2, 1)', 'x', 6, False),
        ('function add_whole() { sum = 0 for (value in _) { sum += value } return(sum) }, x = add_whole(-5, 10, 1.5)', 'x', 6.5, False),
        ('function test(a, b) { return (string(a) + ", " + string(b)) }, x = test(100)', 'x', '100, null', False),
        ('function enclosure(a) {x = a, function closure(y) {return(x + y)}, return(closure)}, z1 = enclosure(100), z2 = enclosure(200), a = z1(5), b = z2(10)', '[a, b]', [105, 210], False),  # Closure
        ('a = ">", b= null, function run(function worker) {a="::", worker()}, run({b = a + " test"})', 'b', ':: test', False),  # Anonymous function in argument.
        ('a = ">", function run(function worker) {worker()}, b= null, a="::", run({b = a + " test"})', 'b', ':: test', False),  # Anonymous function in argument.
        ('a = ">", b= null, function run(function worker) {a="::", worker()}, run() {b = a + " test"}', 'b', ':: test', False),  # Anonymous function in argument.
        ('a = ">", b= null, function run(function worker) {a="::", worker()}, run {b = a + " test"}', 'b', ':: test', False),  # Anonymous function in argument.
        ('a = ">", b= null, function run(function worker) {a="::", worker()}, c = run {b = a + " test"}', 'c', {}, False),  # Anonymous function in argument.
        ('a = ">", b= null, function run(function worker) {a="::", worker(), :=100}, c = run {b = a + " test"}', 'c', 100, False),  # Anonymous function in argument.
        ('a = [1, 2, 3], function x2(before, list, after) {for (i = 0; i < len(list); i += 1) {list[i] *= 2}, return([before, list, after])}, b = x2(a, a, a)', '[a, b]', [[1, 2, 3], [[1, 2, 3], [2, 4, 6], [1, 2, 3]]], False),  # Call by value
        ('a = [1, 2, 3], function x2(before, reference list, after) {for (i = 0; i < len(list); i += 1) {list[i] *= 2}, return([before, list, after])}, b = x2(a, a, a)', '[a, b]', [[2, 4, 6], [[1, 2, 3], [2, 4, 6], [1, 2, 3]]], False),  # Call by reference
        ('a = {"x": 10, "y": 20}, function x2(before, map, after, k) {for (pair in map) {map[pair[0]] = pair[1] * 2}, k *= 2, return([before, map, after, k])}, b = 30, c = x2(a, a, a, b)', '[a, b, c]', [{"x": 10, "y": 20}, 30, [{"x": 10, "y": 20}, {"x": 20, "y": 40}, {"x": 10, "y": 20}, 60]], False),  # Call by value
        ('a = {"x": 10, "y": 20}, function x2(before, reference map, after, reference k) {for (pair in map) {map[pair[0]] = pair[1] * 2}, k *= 2, return([before, map, after, k])}, b = 30, c = x2(a, a, a, b)', '[a, b, c]', [{"x": 20, "y": 40}, 30, [{"x": 10, "y": 20}, {"x": 20, "y": 40}, {"x": 10, "y": 20}, 60]], False),  # Call by reference
        ('function import(a) {return(a*2)}', 'import(3)', None, True),  # "import" is reserved word.
        ('function super(a) {return(a*2)}', 'super(3)', None, True),  # "super" is reserved word.
        ('', 'fake()', None, True),  # fake function is not existed.
    )
)

def test_function(prepared_expression, target_expression, expected_value,
                  is_error):
    """Unit-test for function in Evaluator

    Unit-test for function call and its definition

    Args:
        prepared_expression (str or dict): When the another expression,
                                 such as assignment, is needed before
                                 test, it is given here.
                                 Maybe this is '' mostly.
                                 str: Statements as VivJson
                                 dict: Value as Python's value
        target_expression (str): "return" method's argument as targeted
                                 expression.
        expected_value (Any): The expected value as Python's value
        is_error (bool): True if occurring error is expected,
                         False otherwise.
    """
    config = Config(max_depth=14)

    text = f'return({target_expression})'
    value, error_message = Viv.run(prepared_expression, text, config)
    assert (error_message != '') == is_error

    if isinstance(value, float):
        value = round(value, 10)
    assert value == expected_value

    if error_message == '' \
            and not isinstance(expected_value, bool) \
            and isinstance(expected_value, (int, float)):
        assert type(value) == type(expected_value)  # pylint: disable=C0123

def test_depth():
    """Unit-test for max depth."""
    text = 'function test() {for (i = 0; i < 10; i += 1) { if (i > 5) {return(i)}}}, return(test())'
    value, error_message = Viv.run(text, Config(max_depth=9))
    assert error_message != ''
    assert value is None

    value, error_message = Viv.run(text, Config(max_depth=10))
    assert error_message == ''
    assert value == 6

def test_first_class():
    """Unit-test for 1st class."""
    value, error_message = Viv.run("tests/func_array.viv")
    assert error_message == ''
    assert value == [15, [15, 5, 50, 2], [15, 5, 50, 2], [15, 5, 50, 2], 3, [3, 100, 100.0, "100"], [3, 100, 100.0, "100"], [3, 100, 100.0, "100"]]
